from typing import Literal, Optional

from pydantic import BaseModel, Field, field_validator

from pyravelry.models.base import BaseRavelryModel


class SearchParams(BaseModel):
    """Parameters for the /search.json endpoint.

    [Search Ravelry API documentation](https://www.ravelry.com/api#/_search)
    """

    query: str = Field(...)
    limit: int = Field(50, ge=1, le=500)

    # Using Literal ensures only valid Ravelry types are passed
    types: Optional[
        list[
            Literal[
                "User",
                "PatternAuthor",
                "PatternSource",
                "Pattern",
                "YarnCompany",
                "Yarn",
                "Group",
                "Event",
                "Project",
                "Page",
                "Topic",
                "Shop",
            ]
        ]
    ] = None

    @field_validator("types", mode="before")
    @classmethod
    def format_types(cls, v: str) -> list[str]:
        """Formats the passed in string to be a list.

        Args:
            v (str): space delineated string

        Returns:
            _type_: _description_
        """
        # Allow users to pass a single string instead of a list for convenience
        if isinstance(v, str):
            return v.strip().split(" ")
        return v


class SearchRecordModel(BaseRavelryModel):
    """Details about the specific record found in the search.

    [Search Ravelry API documentation](https://www.ravelry.com/api#/_search)
    """

    type: str
    id: int
    permalink: str
    uri: Optional[str] = None


class SearchResultModel(BaseRavelryModel):
    """Represents an individual result from the global search.

    [Search Ravelry API documentation](https://www.ravelry.com/api#/_search)
    """

    title: str
    type_name: str = Field(..., alias="type_name")
    caption: Optional[str] = None
    tiny_image_url: Optional[str] = None
    image_url: Optional[str] = None
    record: SearchRecordModel


class GlobalSearchResponseModel(BaseRavelryModel):
    """Wrapper for the search.json response.

    [Search Ravelry API documentation](https://www.ravelry.com/api#/_search)
    """

    results: list[SearchResultModel]



from typing import Any

import pytest

from pyravelry.endpoints import FiberAttributesResource
from pyravelry.models import FiberAttributeModel


@pytest.mark.vcr
class TestFiberAttributesResource:
    @pytest.fixture(autouse=True)
    def setup(self, api_info: Any) -> None:
        """Automatically sets up the resource for every test in this class."""
        self.obj = FiberAttributesResource(api_info["client"])

    def test_initialization(self) -> None:
        assert self.obj is not None

    def test_list(self) -> None:
        list_of_fiber_attributes_ = self.obj.list()

        list_of_fiber_attributes = list_of_fiber_attributes_.fiber_attributes

        assert isinstance(list_of_fiber_attributes, list)
        assert len(list_of_fiber_attributes) > 0
        assert isinstance(list_of_fiber_attributes[0], FiberAttributeModel)

        pd_df = list_of_fiber_attributes_.pandas
        pl_df = list_of_fiber_attributes_.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()



from __future__ import annotations

from pydantic import Field

from .base import BaseRavelryModel
from .usersite import UserSiteModel


class PatternAuthorForUserModel(BaseRavelryModel):
    """Represents a PatternAuthor (for_user) object.

    [Pattern Author Ravelry API documentation](https://www.ravelry.com/api#PatternAuthor_result)
    """

    id: int
    name: str = Field(..., description="Designer's full name or alias")
    permalink: str
    favorites_count: int | None = Field(None, description="Number of times the designer has been favorited")
    patterns_count: int | None = Field(None, description="Number of patterns attributed to the designer")


class UserSmallModel(BaseRavelryModel):
    """Represents the User (small) object."""

    id: int
    username: str
    tiny_photo_url: str | None = None
    small_photo_url: str | None = None
    photo_url: str | None = None
    large_photo_url: str | None = None
    profile_country_code: str | None = None


class UserFullModel(UserSmallModel):
    """Represents the User (full/export) object.

    Includes all fields from User (small) plus extended profile information.
    """

    first_name: str | None = None
    about_me: str | None = Field(None, description="Raw 'About me' text")
    about_me_html: str | None = Field(None, description="Rendered HTML 'About me' text")
    fave_colors: str | None = None
    fave_curse: str | None = None
    location: str | None = None

    # Relationships
    pattern_author: PatternAuthorForUserModel | None = Field(None, description="The designer linked to this user")
    user_sites: list[UserSiteModel] | None = Field(None, description="External social/etc sites added to profile")


class UserPostModel(BaseRavelryModel):
    """Model used for updating a User profile via POST.

    [Ravelry API documentation](https://www.ravelry.com/api#User_result)
    """

    first_name: str | None = Field(None, description="User's first name")
    about_me: str | None = Field(None, description='raw "About me" text. Can contain markdown and/or HTML')
    fave_colors: str | None = Field(None, description="User's favorite colors")
    fave_curse: str | None = Field(None, description="User's favorite curse word")
    location: str | None = Field(None, description="User's geographic location")


class UserModel(BaseRavelryModel):
    user: UserFullModel



from pydantic import validate_call

from pyravelry.endpoints.base import BaseEndpoint
from pyravelry.models import UserModel, UserPostModel


class PeopleResource(BaseEndpoint):
    """People endpoint for Ravelry.

    [People Ravelry API documentation](https://www.ravelry.com/api#people_show)
    """

    endpoint: str = "/people"
    output_model = UserModel

    @validate_call
    def show(self, username: str) -> UserModel:
        """
        Get a user's profile information.

        Args:
            username (str): Username or integer ID of the user to lookup.

        Returns:
            UserFullModel: The full user profile data.
        """
        cls = PeopleResource

        url = "/".join([cls.endpoint, f"{username}.json"])

        response_dict = self._fetch(http_client=self._http, endpoint=url, method="GET")

        return cls.output_model.model_validate(response_dict)

    @validate_call
    def update(
        self,
        username: str,
        first_name: str | None = None,
        about_me: str | None = None,
        fave_colors: str | None = None,
        fave_curse: str | None = None,
        location: str | None = None,
    ) -> UserModel:
        """
        Update a user's profile. Requires profile-write permission.

        Args:
            username (str): Username or integer ID of the user to update.
            first_name (str | None): User's first name.
            about_me (str | None): User's "About Me" description.
            fave_colors (str | None): User's favorite colors.
            fave_curse (str | None): User's favorite curse word.
            location (str | None): User's location.

        Returns:
            UserFullModel: The updated user profile.
        """
        cls = PeopleResource

        url = "/".join([cls.endpoint, f"{username}.json"])

        update_data = {
            "first_name": first_name,
            "about_me": about_me,
            "fave_colors": fave_colors,
            "fave_curse": fave_curse,
            "location": location,
        }

        payload = UserPostModel(**update_data).model_dump(exclude_unset=True)

        response_dict = self._fetch(
            http_client=self._http,
            endpoint=url,
            method="POST",
            json=payload,
        )

        return cls.output_model.model_validate(response_dict)



from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class PhotoModel(BaseRavelryModel):
    """Represents a Ravelry Photo object.

    Covers 'Photo' and 'Photo (small)' variants.

    [Photo Ravelry API documentation](https://www.ravelry.com/api#Photo_result)
    """

    id: int
    sort_order: Optional[int] = None
    user_id: Optional[int] = Field(None, description="User ID of the photo's owner/uploader")

    # Captions
    caption: Optional[str] = Field(None, description="Photo caption")
    caption_html: Optional[str] = Field(None, description="Photo caption with image tags")
    copyright_holder: Optional[str] = None

    # Image URLs
    thumbnail_url: Optional[str] = Field(None, description="100px on longest side")
    square_url: Optional[str] = Field(None, description="75x75px square thumbnail")
    small_url: Optional[str] = Field(None, description="240px on longest side")
    small2_url: Optional[str] = Field(None, description="320px on longest side")
    medium_url: Optional[str] = Field(None, description="500px on longest side")
    medium2_url: Optional[str] = Field(None, description="640px on longest side")

    # Offsets
    x_offset: Optional[int] = Field(None, description="X offset for centered photo")
    y_offset: Optional[int] = Field(None, description="Y offset for centered photo")


class PhotoFullModel(PhotoModel):
    """Represents the 'Photo (full)' variant with additional fields.

    [Photo Ravelry API documentation](https://www.ravelry.com/api#Photo_result)
    """

    flickr_url: Optional[str] = None
    shelved_url: Optional[str] = Field(None, description="A book cover image, 150 pixels wide")



from typing import Optional

from pyravelry.models.base import BaseRavelryModel
from pyravelry.models.yarncompany import YarnCompanyModel

from .paginator import PaginatorModel


class YarnCompanySearchParams(BaseRavelryModel):
    """Parameters for the yarn_companies/search endpoint.

    [Yarn Companies Ravelry API documentation](https://www.ravelry.com/api#yarn_companies_search)
    """

    query: Optional[str] = None
    page: int = 1
    page_size: int = 48
    sort: Optional[str] = "best"


class YarnCompanySearchResponseModel(BaseRavelryModel):
    """Response returned by /yarn_companies/search.json.

    [Yarn Companies Ravelry API documentation](https://www.ravelry.com/api#yarn_companies_search)
    """

    yarn_companies: list[YarnCompanyModel]
    paginator: PaginatorModel



from pyravelry.endpoints.base import BaseEndpoint
from pyravelry.models import YarnWeightsModel


class YarnWeightsResource(BaseEndpoint):
    """Endpoint for Color Families.

    Attributes:
        BaseEndpoint (AnyUrl): The endpoint for yarn weights.

    Methods:
        list (list[YarnWeightModel]): returns all yarn weights.

    [Yarn Weights Ravelry API documentation](https://www.ravelry.com/api#/_yarn_weights)
    """

    endpoint: str = "/yarn_weights.json"
    output_model = YarnWeightsModel

    def list(self) -> YarnWeightsModel:
        """
        List the current yarn weights.

        Endpoint: GET /yarn_weights.json
        """
        cls = YarnWeightsResource
        response_dict = self._fetch(http_client=self._http, endpoint=cls.endpoint)
        return cls.output_model.model_validate(response_dict)



from datetime import datetime
from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class ToolModel(BaseRavelryModel):
    """Represents a Ravelry Tool object.

    [Tool Ravelry API documentation](https://www.ravelry.com/api/#Tool_result)
    """

    id: int
    name: Optional[str] = Field(None, description="The name of the tool")
    make: Optional[str] = Field(None, description="The manufacturer/brand of the tool")
    model: Optional[str] = Field(None, description="The specific model of the tool")
    notes: Optional[str] = Field(None, description="User notes regarding the tool")
    created_at: datetime
    updated_at: datetime


class ToolsModel(BaseRavelryModel):
    """Wrapper for the tools.json response.

    [Tool Ravelry API documentation](https://www.ravelry.com/api/#Tool_result)
    """

    tools: list[ToolModel]



from typing import Literal

from pydantic import validate_call

from pyravelry.endpoints.base import BaseEndpoint
from pyravelry.models import (
    CommentFullModel,
    CommentHistoriesModel,
    SimplifiedPaginator,
)


class CommentsResource(BaseEndpoint):
    """Endpoint for Comments.

    Methods:
        create (CommentFullModel): Post a comment related to an object.
        delete (CommentFullModel): Delete a specific comment by ID.
        list (list[CommentHistoryModel]): Get list of comments left by a user.

    [Comments Ravelry API documentation](https://www.ravelry.com/api#/_comments)
    """

    endpoint = "/comments"
    paginator_model = SimplifiedPaginator
    output_model = CommentFullModel
    list_model = CommentHistoriesModel

    @validate_call
    def create(
        self, type_: Literal["project", "pattern", "yarn", "stash"], commented_id: int, body: str, reply_to_id: int
    ) -> CommentFullModel:
        """
        Post a comment related to an object (project, pattern, yarn, or stash).

        Arguments:
            type_ (Literal["project", "pattern", "yarn", "stash"]): The type of item being commented on.
            commented_id (int): ID of the item being commented on.
            body (str): Comment body.
            reply_to_id (int): ID of the comment being replied to. Restricted to item owners.

        Returns:
            (CommentFullModel): The published comment
        """
        cls = CommentsResource

        url = "/".join([cls.endpoint, "create.json"])

        payload = {
            "type": type_,
            "commented_id": commented_id,
            "body": body,
            "reply_to_id": reply_to_id,
        }

        response_dict = self._fetch(http_client=self._http, endpoint=url, method="POST", params=payload)

        return CommentFullModel.model_validate(response_dict["comment"])

    @validate_call
    def delete(self, id_: int) -> CommentFullModel:
        """
        Delete a comment by its ID.

        Arguments:
            id_ (int): The comment ID to delete.

        Returns:
            (CommentFullModel): The deleted comment.
        """
        cls = CommentsResource

        url = "/".join([cls.endpoint, f"{id_}.json"])

        response_dict = self._fetch(http_client=self._http, endpoint=url, method="DELETE")

        return CommentFullModel.model_validate(response_dict["comment"])

    @validate_call
    def list(self, username: str, page: int = 1, page_size: int = 25) -> CommentHistoriesModel:
        """
        Get list of comments left by a specific user.
        """
        cls = CommentsResource

        params = cls.paginator_model(page=page, page_size=page_size)

        url = "/".join(["people", str(username), "comments", "list.json"])
        response_dict = self._fetch(http_client=self._http, endpoint=url, params=params.model_dump())

        data = CommentHistoriesModel.model_validate(response_dict)
        return data



"""Model for YarnWeight type."""

from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class YarnWeightModel(BaseRavelryModel):
    """Represents a Ravelry YarnWeight results object.

    [Yarn Weight Ravelry API documentation](https://www.ravelry.com/api#YarnWeight_result)
    """

    id: int
    name: str

    # These are typically strings or floats representing gauges/thickness
    crochet_gauge: Optional[str] = Field(None, description="Typical crochet gauge")
    knit_gauge: Optional[str] = Field(None, description="Typical knitting gauge")

    max_gauge: Optional[float] = Field(None, description="Maximum recommended gauge")
    min_gauge: Optional[float] = Field(None, description="Minimum recommended gauge")

    ply: Optional[str] = Field(None, description="Number of plies associated with this weight")
    wpi: Optional[str] = Field(None, description="Wraps per inch")


class YarnWeightsModel(BaseRavelryModel):
    """Wrapper for the yarn_weights.json response.

    [Yarn Weight Ravelry API documentation](https://www.ravelry.com/api#YarnWeight_result)
    """

    yarn_weights: list[YarnWeightModel]



from pyravelry.settings import RavelrySettings


class TestRavelrySettings:
    def test_initialization__valid_credentials(self) -> None:
        # this should run without error if
        # RAVELRY_USERNAME and RAVELRY_API_KEY
        # are set on the command line.
        RavelrySettings()



from .base import BaseRavelryModel


class FiberAttributeModel(BaseRavelryModel):
    """Represents a Ravelry FiberAttribute result object.

    [Fiber Attribute Ravelry API documentation](https://www.ravelry.com/api#FiberAttribute_result)
    """

    fiber_attribute_group_id: int
    id: int
    name: str  # Short category name
    permalink: str


class FiberAttributesModel(BaseRavelryModel):
    """Wrapper for the fiber_attributes.json response.

    [Fiber Attribute Ravelry API documentation](https://www.ravelry.com/api#FiberAttribute_result)
    """

    fiber_attributes: list[FiberAttributeModel]



from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class FiberCategoryModel(BaseRavelryModel):
    """Represents a Ravelry FiberCategory results object.

    [Fiber Category Ravelry API documentation](https://www.ravelry.com/api#FiberCategory_result)
    """

    id: int
    name: str = Field(..., description="Short category name")
    permalink: str

    # Optional fields based on the different views of FiberCategory
    short_name: Optional[str] = Field(None, description="Abbreviated category name")

    # parent is listed in attributes but type is often an ID or a nested object
    parent_id: Optional[int] = Field(None, alias="parent", description="Parent category ID")

    # Self-referential list for subcategories
    children: Optional[list["FiberCategoryModel"]] = Field(
        default_factory=list, description="Child categories (subcategories)"
    )


# Required for Pydantic to resolve the recursive "FiberCategoryModel" reference
FiberCategoryModel.model_rebuild()


class FiberCategoriesModel(BaseRavelryModel):
    """Wrapper for the fiber_categories.json response.

    [Fiber Category Ravelry API documentation](https://www.ravelry.com/api#FiberCategory_result)
    """

    fiber_categories: list[FiberCategoryModel]



from typing import Any

import pytest

from pyravelry.endpoints import CommentsResource
from pyravelry.models import CommentFullModel, CommentHistoryModel


@pytest.mark.vcr
class TestSearchResource:
    @pytest.fixture(autouse=True)
    def setup(self, api_info: Any) -> None:
        """Automatically sets up the resource for every test in this class."""
        self.obj = CommentsResource(api_info["client"])

    def test_initialization(self) -> None:
        assert self.obj is not None

    @pytest.mark.dependency(name="comment_create")
    def test_create(self) -> None:
        results = self.obj.create(type_="project", commented_id=12489643, body="This is a reply comment", reply_to_id=1)

        pytest.shared = results.id

        assert isinstance(results, CommentFullModel)

        pd_df = results.pandas
        pl_df = results.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()

    @pytest.mark.dependency(depends=["comment_create"])
    def test_delete(self) -> None:
        results = self.obj.delete(id_=pytest.shared)

        assert isinstance(results, CommentFullModel)

        pd_df = results.pandas
        pl_df = results.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()

    def test_list(self) -> None:
        results = self.obj.list(username="cltheuer")

        data = results.comments

        assert isinstance(data, list)
        assert isinstance(data[0], CommentHistoryModel)

        pd_df = results.pandas
        pl_df = results.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()



"""Yarn provenance models."""

from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class YarnProvenanceModel(BaseRavelryModel):
    """Represents a Ravelry YarnProvenance object.

    This model tracks the origin and processing phases of a yarn.

    [Yarn Provenance Ravelry API documentation](https://www.ravelry.com/api#YarnProvenance_result)
    """

    id: int = Field(..., description="Unique identifier for the provenance record")
    yarn_id: int = Field(..., description="Identifier for the associated yarn")

    # Locations and Descriptions
    country_id: Optional[int] = Field(None, description="Internal ID for the country")
    country_name: Optional[str] = Field(None, description="Name of the country of origin")
    state_id: Optional[int] = Field(None, description="Internal ID for the state/region")
    description: Optional[str] = Field(None, description="Additional details about this provenance phase")

    # Phase Information
    phase_name: Optional[str] = Field(None, description="Name of the production phase (e.g., Dyed, Spun)")
    yarn_phase_id: Optional[int] = Field(None, description="Identifier for the specific yarn phase")


class YarnProvenancesModel(BaseRavelryModel):
    """Wrapper for responses containing multiple yarn provenance records.

    [Yarn Provenance Ravelry API documentation](https://www.ravelry.com/api#YarnProvenance_result)
    """

    yarn_provenances: list[YarnProvenanceModel]



"""Results Objects that are used in endpoints"""

from . import base
from .base import BaseRavelryModel
from .colorfamily import ColorFamiliesModel, ColorFamilyModel
from .comment import (
    CommentCreateModel,
    CommentExportModel,
    CommentFullModel,
    CommentHistoriesModel,
    CommentHistoryModel,
    CommentModel,
)
from .custom_models import paginator
from .custom_models.identifiers import Identifier
from .custom_models.paginator import PaginatorModel, SimplifiedPaginator
from .custom_models.search import (
    GlobalSearchResponseModel,
    SearchParams,
    SearchRecordModel,
    SearchResultModel,
)
from .custom_models.yarncompany import (
    YarnCompanyModel,
    YarnCompanySearchParams,
    YarnCompanySearchResponseModel,
)
from .fiberattribute import FiberAttributeModel, FiberAttributesModel
from .fibercategory import FiberCategoriesModel, FiberCategoryModel
from .socialsite import SocialSiteModel
from .user import UserFullModel, UserModel, UserPostModel, UserSmallModel
from .usersite import UserSiteModel
from .yarnattributegroup import YarnAttributeGroupModel as YarnAttributeModel
from .yarnattributegroup import YarnAttributeGroupsModel as YarnAttributesModel
from .yarncompany import YarnCompaniesModel
from .yarnweight import YarnWeightModel, YarnWeightsModel

__all__ = [
    "BaseRavelryModel",
    "ColorFamiliesModel",
    "ColorFamilyModel",
    "CommentCreateModel",
    "CommentExportModel",
    "CommentFullModel",
    "CommentHistoriesModel",
    "CommentHistoryModel",
    "CommentModel",
    "FiberAttributeModel",
    "FiberAttributesModel",
    "FiberCategoriesModel",
    "FiberCategoryModel",
    "GlobalSearchResponseModel",
    "Identifier",
    "PaginatorModel",
    "SearchParams",
    "SearchRecordModel",
    "SearchResultModel",
    "SimplifiedPaginator",
    "SocialSiteModel",
    "UserFullModel",
    "UserModel",
    "UserPostModel",
    "UserSiteModel",
    "UserSmallModel",
    "YarnAttributeModel",
    "YarnAttributesModel",
    "YarnCompaniesModel",
    "YarnCompanyModel",
    "YarnCompanySearchParams",
    "YarnCompanySearchResponseModel",
    "YarnWeightModel",
    "YarnWeightsModel",
    "base",
    "paginator",
]



from __future__ import annotations

from datetime import date, datetime
from typing import Any

from pydantic import Field

from .base import BaseRavelryModel
from .pack import PackFullModel, PackPostModel

# Assuming these are imported from your other modules
from .photo import PhotoFullModel
from .tool import ToolModel
from .user import UserSmallModel


class ProjectBaseModel(BaseRavelryModel):
    """Base fields shared across most Project variations.

    [Project Ravelry API documentation](https://www.ravelry.com/api#Project_result)
    """

    id: int
    name: str
    permalink: str
    comments_count: int
    favorites_count: int
    photos_count: int
    created_at: datetime
    updated_at: datetime

    craft_id: int | None = None
    craft_name: str | None = None
    progress: int | None = Field(None, ge=0, le=100)
    rating: int | None = None

    started: date | None = None
    started_day_set: bool = True
    completed: date | None = None
    completed_day_set: bool = True

    gauge: float | None = None
    gauge_divisor: int | None = None
    gauge_pattern: str | None = None
    gauge_repeats: int | None = None
    row_gauge: float | None = None

    ends_per_inch: float | None = None
    picks_per_inch: float | None = None

    size: str | None = None
    made_for: str | None = None
    made_for_user_id: int | None = None

    pattern_id: int | None = None
    pattern_name: str | None = None
    status_name: str | None = None
    project_status_id: int | None = None
    project_status_changed: Any | None = None  # Type not specified in docs

    tag_names: list[str] | None = Field(default_factory=list)
    links: dict | None = None


class ProjectSmallModel(ProjectBaseModel):
    """Simplified project view."""

    user_id: int


class ProjectModel(ProjectBaseModel):
    """The standard Project object."""

    user_id: int | None = None
    user: UserSmallModel | None = None
    first_photo: PhotoFullModel | None = None


class ProjectNotesModel(ProjectModel):
    """Project with public notes included."""

    notes: str | None = None
    notes_html: str | None = None


class ProjectFullModel(ProjectNotesModel):
    """Comprehensive project data including packs and tools."""

    packs: list[PackFullModel] | None = Field(default_factory=list)
    photos: list[PhotoFullModel] | None = Field(default_factory=list)
    tools: list[ToolModel] | None = Field(default_factory=list)
    needle_sizes: list[Any] | None = None
    personal_attributes: dict | None = None


class ProjectFullForOwnerModel(ProjectFullModel):
    """Project data with private notes and owner-specific pack info."""

    private_notes: str | None = None
    private_notes_html: str | None = None
    # Overriding packs to use the owner-specific version if necessary
    packs: list[PackFullModel] | None = Field(default_factory=list)


class ProjectPostModel(BaseRavelryModel):
    """Model used for creating or updating a project via POST.

    [Project POST Ravelry API documentation](https://www.ravelry.com/api#Project_update)
    """

    name: str
    craft_id: int | None = None
    made_for: str | None = None
    made_for_user_id: int | None = None
    notes: str | None = None
    private_notes: str | None = None
    progress: int | None = None
    project_status_id: int | None = None
    rating: int | None = None
    started: date | None = None
    completed: date | None = None
    pattern_id: int | None = None
    personal_pattern_name: str | None = None
    size: str | None = None
    tag_names: list[str] | None = None
    needle_sizes: list[int] | None = Field(None, description="List of needle size IDs")
    packs: list[PackPostModel] | None = None


class ProjectsModel(BaseRavelryModel):
    """Wrapper for plural project responses."""

    projects: list[ProjectModel]



from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class YarnAttributeGroupModel(BaseRavelryModel):
    """
    Represents a YarnAttributeGroup.
    This model handles both the 'full' and 'list' versions by making
    the children field optional.

    [Yarn Attribute Group Ravelry API documentation](https://www.ravelry.com/api#YarnAttributeGroup_result)
    """

    id: int
    name: str = Field(..., description="Attribute group name")
    permalink: str

    # Self-referencing field for nested subgroups
    children: Optional[list["YarnAttributeGroupModel"]] = Field(
        default=None, description="Child yarn attribute groups (subgroups)"
    )


# This is required for Pydantic to resolve the recursive "YarnAttributeGroupModel" reference
YarnAttributeGroupModel.model_rebuild()


class YarnAttributeGroupsModel(BaseRavelryModel):
    """Wrapper for responses containing multiple yarn attribute groups.

    [Yarn Attribute Group Ravelry API documentation](https://www.ravelry.com/api#YarnAttributeGroup_result)
    """

    yarn_attribute_groups: list[YarnAttributeGroupModel]



import pandas as pd
import polars as pl
from pydantic import BaseModel, ConfigDict


class BaseRavelryModel(BaseModel):
    """Base Ravelry Results object class"""

    model_config = ConfigDict(
        validate_by_name=True,
        validate_by_alias=True,
        extra="allow",
    )

    @property
    def pandas(self) -> pd.DataFrame:
        """Converts the pyravelry model into a pandas dataframe.

        Returns:
            pd.DataFrame: pandas Dataframe with pyravelry content.
        """
        import pyarrow  # noqa: F401

        data = self.model_dump()
        if len(data.keys()) == 1:
            data = next(data.values().__iter__())
        return pd.json_normalize(data)

    @property
    def polars(self) -> pl.DataFrame:
        """Converts the pyravelry model into a polars dataframe.

        This is very hacky and converts to pandas then polars,
        but the data is sufficiently small to where this doesn't
        matter much.

        Returns:
            pl.DataFrame: polars Dataframe with pyravelry content.
        """
        return pl.from_pandas(self.pandas)



from typing import Any, Optional

from pydantic import Field

from .base import BaseRavelryModel

# Assuming these are implemented in separate files as requested
from .photo import PhotoModel
from .yarnattributegroup import YarnAttributeGroupModel
from .yarncompany import YarnCompanyModel
from .yarnfiber import YarnFiberFullModel
from .yarnprovenance import YarnProvenanceModel
from .yarnweight import YarnWeightModel


class YarnBaseModel(BaseRavelryModel):
    """Base fields shared across all Yarn result types.

    [Yarn Ravelry API documentation](https://www.ravelry.com/api#Yarn_result)
    """

    id: int
    name: str
    permalink: str
    certified_organic: Optional[bool] = None
    discontinued: Optional[bool] = None
    gauge_divisor: Optional[int] = None
    grams: Optional[int] = Field(None, description="weight of a skein/ball of the yarn in grams")
    machine_washable: Optional[bool] = None
    max_gauge: Optional[int] = None
    min_gauge: Optional[int] = None
    rating_average: Optional[float] = None
    rating_count: Optional[int] = None
    rating_total: Optional[int] = None
    texture: Optional[str] = Field(None, description="Obsolete, superseded by yarn attributes")
    thread_size: Optional[Any] = None
    wpi: Optional[int] = Field(None, description="wraps per inch")
    yardage: Optional[int] = None
    yarn_company_name: Optional[str] = None
    organic: Optional[Any] = None


class YarnModel(YarnBaseModel):
    """Represents a Ravelry Yarn result object.

    [Yarn Ravelry API documentation](https://www.ravelry.com/api#Yarn_result)
    """

    first_photo: Optional[PhotoModel] = None
    yarn_weight: Optional[YarnWeightModel] = None


class YarnlistModel(YarnModel):
    """Represents a Ravelry Yarn (list) result object.

    Includes personal_attributes found in search results.

    [Yarn Ravelry API documentation](https://www.ravelry.com/api#Yarn_result)
    """

    personal_attributes: Optional[dict] = Field(None, description="Hash containing favorited, bookmark_id, stash_ids")


class YarnStashlistModel(YarnBaseModel):
    """Represents a Ravelry Yarn (stash_list) result object."""

    yarn_company: Optional[YarnCompanyModel] = None
    yarn_weight: Optional[YarnWeightModel] = None


class YarnFullModel(YarnBaseModel):
    """Represents a Ravelry Yarn (full) result object.

    [Yarn Ravelry API documentation](https://www.ravelry.com/api#Yarn_result)
    """

    max_hook_size: Optional[Any] = None
    max_needle_size: Optional[Any] = None
    min_hook_size: Optional[Any] = None
    min_needle_size: Optional[Any] = None
    notes_html: Optional[str] = None
    personal_attributes: Optional[dict] = None
    photos: list[PhotoModel] = Field(default_factory=list)
    yarn_attributes: list[YarnAttributeGroupModel] = Field(
        default_factory=list, description='Examples: "Superwash", "Gradient"'
    )
    yarn_company: Optional[YarnCompanyModel] = None
    yarn_fibers: list[YarnFiberFullModel] = Field(default_factory=list)
    yarn_provenance: list[YarnProvenanceModel] = Field(default_factory=list, description="Location and processing info")
    yarn_weight: Optional[YarnWeightModel] = None



from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel
from .socialsite import SocialSiteModel


class UserSiteModel(BaseRavelryModel):
    """Represents a Ravelry UserSite object.

    [UserSite Ravelry API documentation](https://www.ravelry.com/api#UserSite_result)
    """

    id: int
    social_site: SocialSiteModel = Field(..., description="Details about the external site")
    url: Optional[str] = Field(None, description="URL to profile page on an external site")
    username: Optional[str] = Field(None, description="Username on an external site")


class UserSitesModel(BaseRavelryModel):
    """Wrapper for a list of user sites, often found in user profile responses.

    [UserSite Ravelry API documentation](https://www.ravelry.com/api#UserSite_result)
    """

    user_sites: list[UserSiteModel]



from __future__ import annotations

from datetime import datetime
from typing import Any, Literal

from pydantic import Field, field_validator

from .base import BaseRavelryModel
from .project import ProjectModel
from .user import UserSmallModel


class CommentCreateModel(BaseRavelryModel):
    """Input parameters for model.

    [Comments Endpoint Ravelry API documentation](https://www.ravelry.com/api/#comments_create)
    """

    type: Literal["project", "pattern", "yarn", "stash"] = Field(
        ..., alias="type", description="The type of item being commented on."
    )
    commented_id: int = Field(..., description="ID of the item being commented on.")
    body: str = Field(..., description="Comment body. Supports Text, Markdown, or HTML.")
    reply_to_id: int | None = Field(None, description="ID of the comment being replied to. Restricted to item owners.")


class CreatedAtBase(BaseRavelryModel):
    """Has an additional created at attribute for handling custom Ravelry parsing.

    [Comments Ravelry API documentation](https://www.ravelry.com/api/#Comment_result)
    """

    created_at: datetime = Field(..., description="Creation time")

    @field_validator("created_at", mode="before")
    @classmethod
    def parse_slashed_datetime(cls, v: Any) -> Any:
        """Parses custom datetime from ravelry."""
        if isinstance(v, str) and "/" in v:
            v = v.replace("/", "-")
            try:
                return datetime.strptime(v, "%Y-%m-%d %H:%M:%S %z")
            except ValueError:
                return v
        return v


class CommentModel(CreatedAtBase):
    """Represents the standard Ravelry Comment object.

    [Comments Ravelry API documentation](https://www.ravelry.com/api/#Comment_result)
    """

    id: int
    comment_html: str = Field(..., description="Comment body (HTML formatted)")
    user: UserSmallModel = Field(..., description="The user who posted the comment")
    highlighted_project: ProjectModel | None = Field(None, description="Related project, if requested and available")


class CommentHistoryModel(CreatedAtBase):
    """Represents a Comment in a history context.

    [Comments Ravelry API documentation](https://www.ravelry.com/api/#Comment_result)
    """

    id: int
    comment_html: str | None = None
    user_id: int = Field(..., description="The ID of the user who posted the comment")
    commentable: Any | None = None  # Polymorphic type depending on context


class CommentHistoriesModel(BaseRavelryModel):
    """Represents multiple Comment in a history context.

    [Comments Ravelry API documentation](https://www.ravelry.com/api/#Comment_result)
    """

    comments: list[CommentHistoryModel]


class CommentFullModel(CommentModel):
    """Represents a full Ravelry Comment with nested replies.

    [Comments Ravelry API documentation](https://www.ravelry.com/api/#Comment_result)
    """

    # Replies are limited to 1 level deep per API docs
    replies: list[CommentModel] = Field(
        default_factory=list, description="Replies to the comment (nesting limited to 1 level)"
    )


class CommentExportModel(CreatedAtBase):
    """Simplified Comment model for export purposes.

    [Comments Ravelry API documentation](https://www.ravelry.com/api/#Comment_result)
    """

    id: int
    comment_html: str



from typing import Optional

from pydantic import Field

from .base import BaseRavelryModel


class ColorFamilyModel(BaseRavelryModel):
    """Represents a Ravelry ColorFamily results object.

    [Color Family Ravelry API documentation](https://www.ravelry.com/api#ColorFamily_result)
    """

    id: int
    name: str
    permalink: str

    # Using Optional because the docs say "Nullable: Yes"
    color: Optional[str] = Field(None, description="HTML color code (e.g., #FFFFFF)")

    spectrum_order: int = Field(..., description="Sort order, (mostly) based on the spectrum")


class ColorFamiliesModel(BaseRavelryModel):
    """Wrapper for the color_families.json response.

    [Color Family Ravelry API documentation](https://www.ravelry.com/api#ColorFamily_result)
    """

    color_families: list[ColorFamilyModel]



from typing import Any, Optional

from pydantic import Field

from .base import BaseRavelryModel


class BasePackModel(BaseRavelryModel):
    """Base fields shared across most Pack response types.

    [Pack Ravelry API documentation](https://www.ravelry.com/api#Pack_result)
    """

    id: int
    color_family_id: Optional[str] = Field(None, description="Color family ID")
    colorway: Optional[str] = None
    dye_lot: Optional[str] = Field(None, description="Dye lot, name or number")
    grams_per_skein: Optional[float] = None
    meters_per_skein: Optional[float] = None
    ounces_per_skein: Optional[float] = None
    personal_name: Optional[str] = Field(None, description="Personal name if not linked to DB yarn")
    prefer_metric_length: Optional[bool] = None
    prefer_metric_weight: Optional[bool] = None
    primary_pack_id: Optional[int] = None
    project_id: Optional[int] = Field(None, description="Project ID affiliated with pack")
    quantity_description: Optional[str] = None
    shop_id: Optional[int] = Field(None, description="Shop ID where purchased")
    shop_name: Optional[str] = None
    skeins: Optional[str] = Field(None, description="Number of skeins")
    stash_id: Optional[int] = Field(None, description="Stash ID affiliated with pack")
    thread_size: Optional[str] = None
    total_grams: Optional[float] = None
    total_meters: Optional[float] = None
    total_ounces: Optional[float] = None
    total_yards: Optional[float] = None
    yards_per_skein: Optional[float] = None
    yarn_id: Optional[int] = Field(None, description="Linked Ravelry yarn ID")


class PackStashModel(BasePackModel):
    """Represents the 'Pack (stash)' and 'Pack (stash_for_owner)' objects.

    [Pack Ravelry API documentation](https://www.ravelry.com/api#Pack_result)
    """

    color_attributes: Optional[Any] = None
    total_paid: Optional[float] = None
    total_paid_currency: Optional[str] = None


class PackFullModel(PackStashModel):
    """Represents the 'Pack (full)' and 'Pack (full_for_owner)' objects.

    [Pack Ravelry API documentation](https://www.ravelry.com/api#Pack_result)
    """

    yarn: Optional[Any] = Field(None, description="Full Yarn object details")
    yarn_name: Optional[str] = None
    yarn_weight: Optional[Any] = None


class PackPublicModel(BaseRavelryModel):
    """Represents the minimal 'Pack (public)' object.

    [Pack Ravelry API documentation](https://www.ravelry.com/api#Pack_result)
    """

    id: int
    yarn_id: Optional[int] = None


class PackPostModel(BaseRavelryModel):
    """Model for creating or updating a Pack (Pack POST).

    [Pack Ravelry API documentation](https://www.ravelry.com/api#Pack_result)
    """

    color_family_id: Optional[str] = None
    colorway: Optional[str] = None
    dye_lot: Optional[str] = None
    length_units: Optional[str] = Field(None, description="One of 'yards', 'meters'")
    personal_gauge_divisor: Optional[str] = None
    personal_max_gauge: Optional[int] = None
    personal_min_gauge: Optional[int] = None
    personal_name: Optional[str] = None
    personal_shop_name: Optional[str] = None
    personal_thread_size: Optional[Any] = None
    personal_yarn_weight_id: Optional[int] = None
    project_id: Optional[int] = None
    purchased_city: Optional[str] = None
    purchased_country_id: Optional[int] = None
    purchased_date: Optional[str] = Field(None, description="YYYY-MM-DD")
    purchased_state_id: Optional[int] = None
    purchased_url: Optional[str] = None
    shop_id: Optional[int] = None
    skein_length: Optional[str] = None
    skein_weight: Optional[str] = None
    skeins: Optional[str] = None
    stash_id: Optional[int] = None
    total_length: Optional[str] = None
    total_paid: Optional[str] = None
    total_paid_currency: Optional[str] = None
    total_weight: Optional[str] = None
    weight_units: Optional[str] = Field(None, description="One of 'grams', 'ounces'")
    yarn_id: Optional[int] = None



from typing import Any, Union

from pydantic import field_validator

from pyravelry.models.base import BaseRavelryModel


class Identifier(BaseRavelryModel):
    """
    Validates that the input is either a string (username)
    or an integer (user ID).

    [Identifier Documentation (URI Parameters)](https://www.ravelry.com/api#people_show)
    """

    id: Union[str, int]

    @field_validator("id")
    @classmethod
    def validate_id(cls, v: Any) -> Any:
        if isinstance(v, str) and not v.strip():
            err = "Username cannot be an empty string"
            raise TypeError(err)
        return v



from typing import Any

import pytest

from pyravelry.endpoints import PeopleResource
from pyravelry.models import UserModel


@pytest.mark.vcr
class TestSearchResource:
    @pytest.fixture(autouse=True)
    def setup(self, api_info: Any) -> None:
        """Automatically sets up the resource for every test in this class."""
        self.obj = PeopleResource(api_info["client"])

    def test_initialization(self) -> None:
        assert self.obj is not None

    def test_show(self) -> None:
        results = self.obj.show(username="cltheuer")

        assert isinstance(results, UserModel)

        pd_df = results.pandas
        pl_df = results.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()

    def test_update_description(self) -> None:
        old_user = self.obj.update(username="cltheuer", fave_colors="Blue")

        new_user = self.obj.update(username="cltheuer", fave_colors="Red")

        assert old_user != new_user

        pd_df = old_user.pandas
        pl_df = old_user.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()



from typing import Any

import pytest

from pyravelry.endpoints import YarnCompaniesResource
from pyravelry.models import YarnCompanyModel, YarnCompanySearchResponseModel


@pytest.mark.vcr
class TestYarnCompanyResources:
    @pytest.fixture(autouse=True)
    def setup(self, api_info: Any) -> None:
        """Automatically sets up the resource for every test in this class."""
        self.obj = YarnCompaniesResource(api_info["client"])

    def test_initialization(self) -> None:
        assert self.obj is not None

    def test_query(self) -> None:
        results = self.obj.query(
            query="Lion",
            page=1,
            page_size=10,
            sort="best_",
        )
        assert isinstance(results, YarnCompanySearchResponseModel)
        assert len(results.yarn_companies) > 0
        assert isinstance(results.yarn_companies[0], YarnCompanyModel)

        pd_df = results.pandas
        pl_df = results.polars

        assert not pd_df.empty
        assert not pl_df.is_empty()



from pyravelry.endpoints.base import BaseEndpoint
from pyravelry.models import ColorFamiliesModel


class ColorFamiliesResource(BaseEndpoint):
    """Endpoint for Color Families.

    Attributes:
        BaseEndpoint (AnyUrl): The endpoint for colorfamily.
        output_model (ColorFamiliesModel): The pydantic model that the api output is validated against.

    Methods:
        list (list[ColorFamilyModel]): returns all color families.

    [Color Family Ravelry API documentation](https://www.ravelry.com/api#/_color_families)
    """

    endpoint: str = "/color_families.json"
    output_model = ColorFamiliesModel

    def list(self) -> ColorFamiliesModel:
        """
        Retrieves all color families from Ravelry.
        """
        cls = ColorFamiliesResource
        response_dict = self._fetch(http_client=self._http, endpoint=cls.endpoint)
        data = cls.output_model.model_validate(response_dict)
        return data
